@@ -59,6 +59,7 @@ class Bancor3(EnvBase):
             },
             whitelisted_tokens=[],
             bootstrapped_tokens=[],
+            exchange_rates={},
             spot_prices={
                 "TKN": {'block_num': [0],
                         'price_usd': [2.50]},
@@ -78,6 +79,7 @@ class Bancor3(EnvBase):
                 "ETH": 0.2,
                 "wBTC": 0.2
             },
+            warning_only=False,
             dao_msig_initialized_pools=[],
             verbose=True,
             **kwargs
@@ -86,9 +88,11 @@ class Bancor3(EnvBase):
         super().__init__()
 
         self.set_param(alpha=alpha,
+                       warning_only=warning_only,
                        spot_prices=spot_prices,
                        vortex_rates=vortex_rates,
                        block_num=block_num,
+                       exchange_rates=exchange_rates,
                        min_liquidity_threshold=min_liquidity_threshold,
                        bnt_funding_limit=bnt_funding_limit,
                        exit_fee=exit_fee,
@@ -128,28 +132,36 @@ class Bancor3(EnvBase):
         else:
             a, b, c, e, m, n, x, w = self.validate_input_types(tkn, x)
 
-            # check surplus
-            self.tokenomics.check_surplus(b, c, e, n)
+            # check balanced
+            self.tokenomics.check_balanced(b, c, e, n)
 
-            # check hlim
-            self.tokenomics.check_hlim(b, c, e, x)
+            if self.tokenomics.is_balanced:
+                case = 'bootstrap surplus'
 
-            # check hmax
-            self.tokenomics.check_hmax(b, c, e, m, n, x, self.tokenomics.surplus)
+            else:
 
-            # check if trading liquidity needs reduced
-            self.tokenomics.check_reduce_trading_liquidity(x, b, c, e, n,
-                                                           self.tokenomics.surplus,
-                                                           self.tokenomics.satisfies_hmax,
-                                                           self.tokenomics.satisfies_hlim)
+                # check surplus
+                self.tokenomics.check_surplus(b, c, e, n)
 
-            # check case (e.g. "default surplus", "arbitrage deficit", etc...)
-            case = self.check_case(self.tokenomics.surplus,
-                                   self.tokenomics.deficit,
-                                   self.tokenomics.satisfies_hlim,
-                                   self.tokenomics.satisfies_hmax,
-                                   self.tokenomics.reduce_trading_liquidity,
-                                   a, b, c, e, w)
+                # check hlim
+                self.tokenomics.check_hlim(b, c, e, x)
+
+                # check hmax
+                self.tokenomics.check_hmax(b, c, e, m, n, x, self.tokenomics.surplus)
+
+                # check if trading liquidity needs reduced
+                self.tokenomics.check_reduce_trading_liquidity(x, b, c, e, n,
+                                                               self.tokenomics.surplus,
+                                                               self.tokenomics.satisfies_hmax,
+                                                               self.tokenomics.satisfies_hlim)
+
+                # check case (e.g. "default surplus", "arbitrage deficit", etc...)
+                case = self.check_case(self.tokenomics.surplus,
+                                       self.tokenomics.deficit,
+                                       self.tokenomics.satisfies_hlim,
+                                       self.tokenomics.satisfies_hmax,
+                                       self.tokenomics.reduce_trading_liquidity,
+                                       a, b, c, e, w)
 
             if case in ['bootstrap surplus', 'bootstrap surplus (special case)']:
                 s = self.tokenomics.handle_bootstrap_surplus(x, n)
@@ -188,7 +200,7 @@ class Bancor3(EnvBase):
             tkn_out = s + u
             bnt_out = t
 
-        self.check_pool_shutdown()
+        self.check_pool_shutdown(warning_only=self.warning_only)
         return tkn_out, bnt_out
 
     def stake(self, tkn, x, block_num):
@@ -217,10 +229,10 @@ class Bancor3(EnvBase):
                 self.add_tkn_to_available_liquidity_ledger(tkn, x)
             self.check_bootstrapped_tokens()
 
-        self.check_pool_shutdown()
+        self.check_pool_shutdown(warning_only=self.warning_only)
         return (y)
 
-    def trade(self, tkn1, x, tkn2, block_num):
+    def trade(self, tkn1, x, tkn2, block_num, update_liquidity=False):
         """Main protocol logic to perform trading functionality.
 
         Args:
@@ -253,7 +265,7 @@ class Bancor3(EnvBase):
             bnt_trading_liquidity, tkn_trading_liquidity, tkn_out, tkn_fee, vortex_fee = operation(
                 a, d, b, x, e)
 
-            if self.is_within_ema_tolerance(tkn_a):
+            if self.is_within_ema_tolerance(tkn_a) or update_liquidity:
                 self.update_available_liquidity_ledger(tkn_a, bnt_trading_liquidity, tkn_trading_liquidity)
 
             self.add_tkn_to_vault(tkn1, x)
@@ -282,7 +294,6 @@ class Bancor3(EnvBase):
             bnt_source_trading_liquidity, tkn_source_trading_liquidity, bnt_destination_trading_liquidity, tkn_destination_trading_liquidity, tkn_out, bnt_fee, tkn_fee, vortex_fee = self.tokenomics.handle_trade_tkn_to_tkn(
                 a1, d1, b1, a2, d2, b2, x, e)
 
-
             if self.is_within_ema_tolerance(tkn1):
                 self.update_available_liquidity_ledger(tkn1, bnt_source_trading_liquidity, tkn_source_trading_liquidity)
 
@@ -304,8 +315,10 @@ class Bancor3(EnvBase):
                   f'tkn_fee={tkn_fee}, \n',
                   f'vortex_fee={vortex_fee} \n')
 
-        self.check_pool_shutdown()
-        # self.get_spot_rate(tkn1, update_spot_prices=True)
-        # self.get_spot_rate(tkn2, update_spot_prices=True)
+        self.check_pool_shutdown(warning_only=self.warning_only)
+
+        for token in [tkn1, tkn2]:
+            self.update_spot_prices(token)
+
         return tkn_out
 
