@@ -59,6 +59,15 @@ class EnvBase(ABC):
     def min_liquidity_threshold(self, updated_min_liquidity_threshold: int or float):
         self._min_liquidity_threshold = updated_min_liquidity_threshold
 
+    @property
+    def warning_only(self) -> bool:
+        """If True then will cause exception and break whenever a negative balance is found."""
+        return self._warning_only
+
+    @warning_only.setter
+    def warning_only(self, updated_warning_only: bool):
+        self._warning_only = updated_warning_only
+
     @property
     def bnt_funding_limit(self) -> int or float:
         """The BancorDAO determines the available liquidity for trading, through adjustment of the “BNT funding limit” parameter."""
@@ -179,13 +188,6 @@ class EnvBase(ABC):
     @spot_prices.setter
     def spot_prices(self, updated_spot_prices: dict):
         self._spot_prices = updated_spot_prices
-        exchange_rates = {}
-        bnt_price = self._spot_prices['BNT']['price_usd'][-1]
-        for tkn in self._spot_prices:
-            tkn_price = self._spot_prices[tkn]['price_usd'][-1]
-            exchange_rates[tkn] = tkn_price / bnt_price
-        self.exchange_rates = exchange_rates
-        self.set_whitelisted_tokens(whitelisted_tokens=[tkn for tkn in updated_spot_prices])
 
     @property
     def custom_name(self):
@@ -208,9 +210,11 @@ class EnvBase(ABC):
                         spot_prices=None,
                         whitelisted_tokens=None,
                         bootstrapped_tokens=None,
+                        exchange_rates=None,
                         external_price_feed=None,
                         vortex_rates=None,
                         dao_msig_initialized_pools=None,
+                        warning_only=None,
                         verbose=None
                   ):
         """Allows user to set the specified system parameters.
@@ -233,12 +237,16 @@ class EnvBase(ABC):
         """
         if block_num is not None:
             self.block_num = block_num
+        if exchange_rates is not None:
+            self.exchange_rates = exchange_rates
         if min_liquidity_threshold is not None:
             self.min_liquidity_threshold = min_liquidity_threshold
         if bnt_funding_limit is not None:
             self.bnt_funding_limit = bnt_funding_limit
         if alpha is not None:
             self.alpha = alpha
+        if warning_only is not None:
+            self.warning_only = warning_only
         if cooldown_period is not None:
             self.cooldown_period = cooldown_period
         if pool_fees is not None:
@@ -246,7 +254,7 @@ class EnvBase(ABC):
         if exit_fee is not None:
             self.exit_fee = exit_fee
         if whitelisted_tokens is not None:
-            self.whitelisted_tokens = whitelisted_tokens
+            self.set_whitelisted_tokens(whitelisted_tokens)
         if vortex_rates is not None:
             self.vortex_rates = vortex_rates
         if external_price_feed is not None:
@@ -262,10 +270,25 @@ class EnvBase(ABC):
             self.tokenomics = FixedPointTokenomics() if is_solidity else FloatingPointTokenomics()
         if spot_prices is not None:
             self.spot_prices = spot_prices
-            for tkn in spot_prices:
-                if tkn in self.ema:
-                    self.ema[tkn]['block_num'] = [self.block_num]
-                    self.ema[tkn]['ema'] = [self.exchange_rates[tkn]]
+
+        self.update_exchange_rates()
+        self.init_ema()
+
+    def init_ema(self):
+        ema = self.ema
+        for tkn in self.spot_prices:
+            if tkn in ema:
+                ema[tkn]['block_num'] = [self.block_num]
+                ema[tkn]['ema'] = [self.exchange_rates[tkn]]
+        self.ema = ema
+
+    def update_exchange_rates(self):
+        exchange_rates = self.exchange_rates
+        bnt_price = self.spot_prices['BNT']['price_usd'][-1]
+        for tkn in self.spot_prices:
+            tkn_price = self.spot_prices[tkn]['price_usd'][-1]
+            exchange_rates[tkn] = tkn_price / bnt_price
+        self.exchange_rates = exchange_rates
 
     def get_available_liquidity_ledger(self, tkn):
         """Gets the available trading liquidity ledger.
@@ -322,11 +345,26 @@ class EnvBase(ABC):
         staked_amount = self.staking_ledger[tkn][-1]
         bnTKN_supply = self.pool_token_supply_ledger[f'bn{tkn}_ERC20_contract']['supply'][-1]
         if staked_amount > 0 and bnTKN_supply > 0:
-            bnTKN_value = bnTKN_supply / staked_amount
+            bnTKN_value = float(bnTKN_supply) / float(staked_amount)
         else:
             bnTKN_value = 1
         return bnTKN_value
 
+    def update_spot_prices(self, tkn):
+        """Updates the internal spot price as determined by the trading liquidity balances of the pool
+
+        Args:
+            tkn (str): The value of BNT to add to the vortex
+        Returns:
+            self
+        """
+        spot_rate = self.get_spot_rate(tkn)
+        spot_prices = self.spot_prices
+        spot_prices[tkn]['block_num'].append(self.block_num)
+        spot_prices[tkn]['price_usd'].append(spot_rate)
+        self.spot_prices = spot_prices
+        self.update_exchange_rates()
+
     def get_spot_rate(self, tkn, update_spot_prices=False):
         """Gets the spot rate in units of BNT/TKN as determined by the trading liquidity balances of the pool
 
@@ -350,6 +388,7 @@ class EnvBase(ABC):
             spot_prices[tkn]['block_num'].append(self.block_num)
             spot_prices[tkn]['price_usd'].append(spot_rate)
             self.spot_prices = spot_prices
+            self.update_exchange_rates()
         return spot_rate
 
     def get_bnt_rate(self, tkn) -> float:
@@ -426,12 +465,6 @@ class EnvBase(ABC):
             self
         """
         ema = self.ema
-        for token in ema:
-            l = ema[token]['block_num'] # gets list of the current block history for any token
-            break
-
-        # ema[tkn] = {'block_num': [x for x in l], 'ema': [self.exchange_rates[tkn] for x in range(len(l))]}
-
         if tkn in self.ema:
             if ema[tkn]['block_num'][-1] < self.block_num:
                 ema[tkn]['block_num'].append(self.block_num)
@@ -504,6 +537,14 @@ class EnvBase(ABC):
         self.vortex_ledger['BNT'].append(self.vortex_ledger['BNT'][-1] + x)
         self.vortex_ledger['block_num'].append(self.block_num)
 
+    def add_tkn_to_pool_token_supply_ledger(self, tkn, x):
+        self.pool_token_supply_ledger[tkn]['block_num'].append(self.block_num)
+        for key in self.pool_token_supply_ledger:
+            if key == tkn:
+                self.pool_token_supply_ledger[tkn]['supply'].append(self.pool_token_supply_ledger[tkn]['supply'][-1] + x)
+            else:
+                self.pool_token_supply_ledger[key]['supply'].append(self.pool_token_supply_ledger[key]['supply'][-1])
+
     def add_tkn_to_available_liquidity_ledger(self, tkn, x):
         """Adds TKN to the available liquidity ledger.
 
@@ -516,19 +557,49 @@ class EnvBase(ABC):
         """
         for key in self.available_liquidity_ledger:
             if key == tkn:
-                if self.available_liquidity_ledger[tkn]['funding_limit'][-1] > 0:
-                    pool_token_value = self.get_pool_token_value(key)
-                    funding_limit = (self.available_liquidity_ledger[key]['BNT'][-1] + self.available_liquidity_ledger[tkn]['funding_limit'][-1])
-                    max_limit = (self.available_liquidity_ledger[key]['BNT'][-1] * 2)
-                    funding_amt = max_limit if max_limit < funding_limit else funding_limit
-                    self.available_liquidity_ledger[key][tkn].append(self.available_liquidity_ledger[key][key][-1] + (funding_amt / 2 ) / self.get_spot_rate(key))
-                    self.available_liquidity_ledger[key]['BNT'].append(funding_amt)
-                    self.available_liquidity_ledger[key]['block_num'].append(self.block_num)
-                    self.available_liquidity_ledger[key]['funding_limit'].append(
-                        self.available_liquidity_ledger[tkn]['funding_limit'][-1] - funding_amt)
-                    self.vault_ledger['BNT'][-1] += funding_amt / 2
-                    self.staking_ledger['BNT'][-1] += funding_amt / 2
-                    self.pool_token_supply_ledger[f"bnBNT_ERC20_contract"]["supply"][-1] += funding_amt / 2
+
+                a = self.available_liquidity_ledger[key]['BNT'][-1]
+                b = self.available_liquidity_ledger[key][tkn][-1]
+                c = self.vault_ledger[key][-1] - b
+                spot_rate = self.get_spot_rate(tkn)
+                funding_limit = self.available_liquidity_ledger[key]['funding_limit'][-1]
+
+                if (c >= b) & (a <= funding_limit):
+                    self.available_liquidity_ledger[key]['BNT'].append(a * 2)
+                    self.available_liquidity_ledger[key][tkn].append(b * 2)
+                    self.available_liquidity_ledger[key]['funding_limit'].append(funding_limit - a)
+
+                elif (c >= b) & (a > funding_limit):
+                    self.available_liquidity_ledger[key]['BNT'].append(a + funding_limit)
+                    self.available_liquidity_ledger[key][tkn].append(funding_limit / spot_rate)
+                    self.available_liquidity_ledger[key]['funding_limit'].append(0)
+
+                elif (c < b) & (a <= funding_limit):
+                    self.available_liquidity_ledger[key]['BNT'].append((b + c) * spot_rate)
+                    self.available_liquidity_ledger[key][tkn].append(b + c)
+                    self.available_liquidity_ledger[key]['funding_limit'].append(funding_limit - (c * spot_rate))
+
+                else:
+                    self.available_liquidity_ledger[key]['BNT'].append(a + (c * spot_rate))
+                    self.available_liquidity_ledger[key][tkn].append(b + c)
+                    self.available_liquidity_ledger[key]['funding_limit'].append(funding_limit - (c * spot_rate))
+
+                self.available_liquidity_ledger[key]['block_num'].append(self.block_num)
+                final_liquidity = self.available_liquidity_ledger[key]['BNT'][-1]
+                delta = final_liquidity - a
+                self.add_tkn_to_vault('BNT', delta)
+                self.add_tkn_to_staking_ledger('BNT', delta)
+                bnt_bnbnt_rate = self.staking_ledger['BNT'][-1] / self.pool_token_supply_ledger[f"bnBNT_ERC20_contract"]["supply"][-1]
+                update_amt = delta / bnt_bnbnt_rate
+                self.add_tkn_to_pool_token_supply_ledger("bnBNT_ERC20_contract", update_amt)
+
+                if "bnBNT" not in self.protocol_agent.wallet:
+                    self.protocol_agent.wallet["bnBNT"] = {'block_num':[0],
+                                                           "bnBNT": [0]}
+
+                self.protocol_agent.wallet["bnBNT"]["bnBNT"].append(self.protocol_agent.wallet["bnBNT"]["bnBNT"][-1] + update_amt)
+                self.protocol_agent.wallet["bnBNT"]['block_num'].append(self.block_num)
+
             else:
                 self.available_liquidity_ledger[key][key].append(self.available_liquidity_ledger[key][key][-1])
                 self.available_liquidity_ledger[key]['BNT'].append(self.available_liquidity_ledger[key]['BNT'][-1])
@@ -546,8 +617,8 @@ class EnvBase(ABC):
         Returns:
             y (int): The number of pool tokens issued to the user.
         """
-        if bnt is None: y = x * self.get_pool_token_value(tkn)
-        else: y = bnt
+        if bnt is None: y = float(x) * float(self.get_pool_token_value(tkn))
+        else: y = float(bnt)
         for contract in self.pool_token_supply_ledger:
             self.pool_token_supply_ledger[contract]['block_num'].append(self.block_num)
             if contract == f'bn{tkn}_ERC20_contract':
@@ -565,16 +636,22 @@ class EnvBase(ABC):
         dao_msig_initialized_pools = self.dao_msig_initialized_pools
         for TKN in self.whitelisted_tokens:
             BNTperTKN = self.get_bnt_rate(TKN)
-            if float(self.vault_ledger[TKN][-1]) * float(BNTperTKN) < self.min_liquidity_threshold and (TKN in self.dao_msig_initialized_pools):
-                # if not warning_only:
-                #     dao_msig_initialized_pools = [token for token in dao_msig_initialized_pools if token != TKN]
-                #     self.available_liquidity_ledger[TKN]['BNT'].append(0)
-                #     self.available_liquidity_ledger[TKN][TKN].append(0)
-                #     self.available_liquidity_ledger[TKN]['funding_limit'].append(self.bnt_funding_limit)
-                #     self.available_liquidity_ledger[TKN]['block_num'].append(
-                #         self.available_liquidity_ledger[TKN]['block_num'][-1])
-                #     self.dao_msig_initialized_pools = dao_msig_initialized_pools
+            try:
+                if self.vault_ledger[TKN][-1] < 0 or self.available_liquidity_ledger[TKN][-1] < 0 or self.staking_ledger[TKN][-1] < 0 or self.external_protection_wallet_ledger[TKN][-1] < 0:
+                    if not warning_only:
+                        raise UserWarning(f'Error. {TKN} balance cannot be less than zero. {self.vault_ledger[TKN][-1]}')
+                    #     dao_msig_initialized_pools = [token for token in dao_msig_initialized_pools if token != TKN]
+                    #     self.available_liquidity_ledger[TKN]['BNT'].append(0)
+                    #     self.available_liquidity_ledger[TKN][TKN].append(0)
+                    #     self.available_liquidity_ledger[TKN]['funding_limit'].append(self.bnt_funding_limit)
+                    #     self.available_liquidity_ledger[TKN]['block_num'].append(
+                    #         self.available_liquidity_ledger[TKN]['block_num'][-1])
+                    #     self.dao_msig_initialized_pools = dao_msig_initialized_pools
+            except:
+                pass
 
+            if float(self.vault_ledger[TKN][-1]) * float(BNTperTKN) < self.min_liquidity_threshold and (
+                        TKN in self.dao_msig_initialized_pools):
                 if self.verbose:
                     print(
                         f"WARNING POOL SHUTDOWN!!! \n",
@@ -699,6 +776,7 @@ class EnvBase(ABC):
         Returns:
             self
         """
+        # print('tkn_a=', tkn, 'bnt_trading_liquidity=', bnt_trading_liquidity)
         self.available_liquidity_ledger[tkn]['BNT'].append(bnt_trading_liquidity)
         self.available_liquidity_ledger[tkn][tkn].append(tkn_trading_liquidity)
         self.available_liquidity_ledger[tkn]['funding_limit'].append(
@@ -733,25 +811,25 @@ class EnvBase(ABC):
 
             # handle surplus signs
             if (case != 'default surplus') & self.tokenomics.surplus:
-                self.available_liquidity_ledger[tkn]['BNT'][-1] -= p
-                self.available_liquidity_ledger[tkn][tkn][-1] += r
+                self.available_liquidity_ledger[tkn]['BNT'][-1] -= float(p)
+                self.available_liquidity_ledger[tkn][tkn][-1] += float(r)
             elif self.tokenomics.surplus:
-                self.available_liquidity_ledger[tkn]['BNT'][-1] -= p
-                self.available_liquidity_ledger[tkn][tkn][-1] -= r
+                self.available_liquidity_ledger[tkn]['BNT'][-1] -= float(p)
+                self.available_liquidity_ledger[tkn][tkn][-1] -= float(r)
 
             # handle deficit signs
             elif (self.tokenomics.deficit) & ('ep2' in case):
-                self.available_liquidity_ledger[tkn]['BNT'][-1] -= p
-                self.available_liquidity_ledger[tkn][tkn][-1] -= r
+                self.available_liquidity_ledger[tkn]['BNT'][-1] -= float(p)
+                self.available_liquidity_ledger[tkn][tkn][-1] -= float(r)
             elif self.tokenomics.deficit:
-                self.available_liquidity_ledger[tkn]['BNT'][-1] += p
-                self.available_liquidity_ledger[tkn][tkn][-1] -= r
+                self.available_liquidity_ledger[tkn]['BNT'][-1] += float(p)
+                self.available_liquidity_ledger[tkn][tkn][-1] -= float(r)
 
-            self.staking_ledger['BNT'][-1] -= q
-            self.staking_ledger[tkn][-1] -= x
-            self.vault_ledger[tkn][-1] -= s
+            self.staking_ledger['BNT'][-1] -= float(q)
+            self.staking_ledger[tkn][-1] -= float(x)
+            self.vault_ledger[tkn][-1] -= float(s)
             self.mint_bntkn(tkn, bnt=t)
-            self.external_protection_wallet_ledger[tkn][-1] -= u
+            self.external_protection_wallet_ledger[tkn][-1] -= float(u)
 
         eq_case = self.tokenomics.reduce_trading_liquidity
         if self.tokenomics.surplus: eq = 'x(1-n) <= c'
@@ -837,41 +915,79 @@ class EnvBase(ABC):
         if withdraw:
             logger = self.describe_withdrawal
         else:
-            try:
-                logger = {
-                    'trading_liquidity':[f"BNT:{self.available_liquidity_ledger['ETH']['BNT'][-1]} - ETH:{self.available_liquidity_ledger['ETH']['ETH'][-1]}",
-                                         f"BNT:{self.available_liquidity_ledger['wBTC']['BNT'][-1]} - wBTC:{self.available_liquidity_ledger['wBTC']['wBTC'][-1]}",
-                                         f"BNT:{self.available_liquidity_ledger['LINK']['BNT'][-1]} - LINK:{self.available_liquidity_ledger['LINK']['LINK'][-1]}",
-                                         ''],
-                    'vault_ledger': [
-                        f"BNT:{self.vault_ledger['BNT'][-1]}",
-                        f"ETH:{self.vault_ledger['ETH'][-1]}",
-                        f"wBTC:{self.vault_ledger['wBTC'][-1]}",
-                        f"LINK:{self.vault_ledger['LINK'][-1]}",
-                    ],
-                    'staking_ledger': [
-                        f"BNT:{self.staking_ledger['BNT'][-1]}",
-                        f"ETH:{self.staking_ledger['ETH'][-1]}",
-                        f"wBTC:{self.staking_ledger['wBTC'][-1]}",
-                        f"LINK:{self.staking_ledger['LINK'][-1]}",
-                    ],
-                    'pool_token_supply_ledger': [
-                        f"bnBNT:{self.pool_token_supply_ledger['bnBNT_ERC20_contract']['supply'][-1]}",
-                        f"bnETH:{self.pool_token_supply_ledger['bnETH_ERC20_contract']['supply'][-1]}",
-                        f"bnwBTC:{self.pool_token_supply_ledger['bnwBTC_ERC20_contract']['supply'][-1]}",
-                        f"bnLINK:{self.pool_token_supply_ledger['bnLINK_ERC20_contract']['supply'][-1]}",
-                    ],
-                    'vortex_ledger': [
-                        f"Balance:{self.vortex_ledger['BNT'][-1]}",
-                        '',
-                        '',
-                        '',
-                    ],
-                }
-            except:
-                try:
-                    logger = self.describe_withdrawal
-                except:
-                    raise UserWarning('Check ledger contents individually to ensure all required tokens are found.')
+            # try:
+            logger = {'trading_liquidity':[],
+                      'vault_ledger': [],
+                      'staking_ledger': [],
+                      'pool_token_supply_ledger': [],
+                      'vortex_ledger': []
+                      }
+            for tkn in self.available_liquidity_ledger:
+                if tkn != 'BNT':
+                    logger['trading_liquidity'].append(f"{tkn}:{self.available_liquidity_ledger[tkn][tkn][-1]} - BNT:{self.available_liquidity_ledger[tkn]['BNT'][-1]}")
+            for tkn in self.available_liquidity_ledger:
+                logger['vault_ledger'].append(
+                    f"{tkn}:{self.vault_ledger[tkn][-1]}")
+            for tkn in self.staking_ledger:
+                if tkn != 'block_num':
+                    logger['staking_ledger'].append(
+                    f"{tkn}:{self.staking_ledger[tkn][-1]}")
+            for tkn in self.pool_token_supply_ledger:
+                logger['pool_token_supply_ledger'].append(
+                    f"{tkn.replace('_ERC20_contract','')}:{self.pool_token_supply_ledger[tkn]['supply'][-1]}")
+
+            logger['vortex_ledger'].append(f"Balance:{self.vortex_ledger['BNT'][-1]}")
+
+            max_rows = max([len(logger['pool_token_supply_ledger']),
+                            len(logger['staking_ledger']),
+                            len(logger['vault_ledger']),
+                            len(logger['trading_liquidity'])])
+
+
+
+            for key in logger:
+                for row in range(max_rows - len(logger[key])):
+                    logger[key].append('')
+
+            for ledger in logger:
+                logger[ledger].sort()
+
+                # logger = {
+                #     'trading_liquidity':[f"BNT:{self.available_liquidity_ledger['ETH']['BNT'][-1]} - ETH:{self.available_liquidity_ledger['ETH']['ETH'][-1]}",
+                #                          f"BNT:{self.available_liquidity_ledger['wBTC']['BNT'][-1]} - wBTC:{self.available_liquidity_ledger['wBTC']['wBTC'][-1]}",
+                #                          f"BNT:{self.available_liquidity_ledger['LINK']['BNT'][-1]} - LINK:{self.available_liquidity_ledger['LINK']['LINK'][-1]}",
+                #                          ''],
+                #     'vault_ledger': [
+                #         f"BNT:{self.vault_ledger['BNT'][-1]}",
+                #         f"ETH:{self.vault_ledger['ETH'][-1]}",
+                #         f"wBTC:{self.vault_ledger['wBTC'][-1]}",
+                #         f"LINK:{self.vault_ledger['LINK'][-1]}",
+                #     ],
+                #     'staking_ledger': [
+                #         f"BNT:{self.staking_ledger['BNT'][-1]}",
+                #         f"ETH:{self.staking_ledger['ETH'][-1]}",
+                #         f"wBTC:{self.staking_ledger['wBTC'][-1]}",
+                #         f"LINK:{self.staking_ledger['LINK'][-1]}",
+                #     ],
+                #     'pool_token_supply_ledger': [
+                #         f"bnBNT:{self.pool_token_supply_ledger['bnBNT_ERC20_contract']['supply'][-1]}",
+                #         f"bnETH:{self.pool_token_supply_ledger['bnETH_ERC20_contract']['supply'][-1]}",
+                #         f"bnwBTC:{self.pool_token_supply_ledger['bnwBTC_ERC20_contract']['supply'][-1]}",
+                #         f"bnLINK:{self.pool_token_supply_ledger['bnLINK_ERC20_contract']['supply'][-1]}",
+                #     ],
+                #     'vortex_ledger': [
+                #         f"Balance:{self.vortex_ledger['BNT'][-1]}",
+                #         '',
+                #         '',
+                #         '',
+                #     ],
+                # }
+
+            # except:
+            #     try:
+            #         logger = self.describe_withdrawal
+            #     except:
+            #         raise UserWarning('Check ledger contents individually to ensure all required tokens are found.')
+            #
         return pd.DataFrame(logger)
 
